/* BotServ functions
 *
 * Epona (c) 2000 PegSoft
 * Contact us at epona@pegirc.com
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */
 
/*************************************************************************/
 
#include "services.h"
#include "pseudo.h"

/*************************************************************************/

BotInfo *botlists[256];			/* Hash list of bots */
int nbots = 0;

/*************************************************************************/

static BotInfo *makebot(char *nick);
static UserData *get_user_data(Channel *c, User *u);
static void unassign(User *u, ChannelInfo *ci);

static void check_ban(ChannelInfo *ci, User *u, int ttbtype);
static void bot_kick(ChannelInfo *ci, User *u, int message, ...);
static void bot_raw_ban(ChannelInfo *ci, char *nick);
static void bot_raw_kick(ChannelInfo *ci, char *nick, char *reason);
static void bot_raw_mode(ChannelInfo *ci, char *mode, char *nick);
static void bot_raw_unban(ChannelInfo *ci, char *nick);

static void do_help(User *u);
static void do_bot(User *u);
static void do_botlist(User *u);
static void do_assign(User *u);
static void do_unassign(User *u);
static void do_info(User *u);
static void do_set(User *u);
static void do_kickcmd(User *u);
static void do_badwords(User *u);
static void do_say(User *u);
static void do_act(User *u);

/*************************************************************************/

static Command cmds[] = {
    { "HELP",			    do_help,     NULL,  -1,                          -1,-1,-1,-1 },
    { "BOTLIST",		    do_botlist,  NULL,  BOT_HELP_BOTLIST,            -1,-1,-1,-1 },
    { "ASSIGN",			    do_assign,   NULL,  BOT_HELP_ASSIGN,             -1,-1,-1,-1 },
    { "UNASSIGN",		    do_unassign, NULL,  BOT_HELP_UNASSIGN,           -1,-1,-1,-1 },
    { "INFO",               do_info,     NULL,  BOT_HELP_INFO,               -1,-1,-1,-1 },
    { "SET",                do_set,      NULL,  BOT_HELP_SET,                
        -1, BOT_SERVADMIN_HELP_SET, 
        BOT_SERVADMIN_HELP_SET, BOT_SERVADMIN_HELP_SET },
    { "SET DONTKICKOPS",    NULL,        NULL,  BOT_HELP_SET_DONTKICKOPS,    -1,-1,-1,-1 },
    { "SET DONTKICKVOICES", NULL,        NULL,  BOT_HELP_SET_DONTKICKVOICES, -1,-1,-1,-1 },
    { "SET FANTASY",		NULL,		 NULL,	BOT_HELP_SET_FANTASY,		 -1,-1,-1,-1 },
    { "SET GREET",			NULL,		 NULL,	BOT_HELP_SET_GREET,		 	 -1,-1,-1,-1 },
    { "SET SYMBIOSIS",		NULL,		 NULL,	BOT_HELP_SET_SYMBIOSIS,		 -1,-1,-1,-1 },
    { "KICK",               do_kickcmd,  NULL,  BOT_HELP_KICK,               -1,-1,-1,-1 },
    { "KICK BADWORDS",      NULL,        NULL,  BOT_HELP_KICK_BADWORDS,      -1,-1,-1,-1 },
    { "KICK BOLDS",         NULL,        NULL,  BOT_HELP_KICK_BOLDS,         -1,-1,-1,-1 },
    { "KICK CAPS",          NULL,        NULL,  BOT_HELP_KICK_CAPS,          -1,-1,-1,-1 },
    { "KICK COLORS",        NULL,        NULL,  BOT_HELP_KICK_COLORS,        -1,-1,-1,-1 },
    { "KICK FLOOD",         NULL,        NULL,  BOT_HELP_KICK_FLOOD,         -1,-1,-1,-1 },
    { "KICK REPEAT",        NULL,        NULL,  BOT_HELP_KICK_REPEAT,        -1,-1,-1,-1 },
    { "KICK REVERSES",      NULL,        NULL,  BOT_HELP_KICK_REVERSES,      -1,-1,-1,-1 },
    { "KICK UNDERLINES",    NULL,        NULL,  BOT_HELP_KICK_UNDERLINES,    -1,-1,-1,-1 },
    { "BADWORDS",           do_badwords, NULL,  BOT_HELP_BADWORDS,           -1,-1,-1,-1 },
    { "SAY",				do_say,		 NULL,  BOT_HELP_SAY,				 -1,-1,-1,-1 },
    { "ACT",				do_act,		 NULL,  BOT_HELP_ACT,				 -1,-1,-1,-1 },
    
    /* Services admins commands */
    { "BOT",       do_bot,   is_services_admin,  -1,
		           -1, BOT_SERVADMIN_HELP_BOT,
		           BOT_SERVADMIN_HELP_BOT, BOT_SERVADMIN_HELP_BOT },
	{ "SET NOBOT", NULL,	 NULL,	-1,		     
	               -1, BOT_SERVADMIN_HELP_SET_NOBOT,
	               BOT_SERVADMIN_HELP_SET_NOBOT, BOT_SERVADMIN_HELP_SET_NOBOT },
		
    { NULL }
};

/*************************************************************************/
/*************************************************************************/

/* Return information on memory use.  Assumes pointers are valid. */

void get_botserv_stats(long *nrec, long *memuse)
{
    long count = 0, mem = 0;
    int i;
    BotInfo *bi;

	for (i = 0; i < 256; i++) {
    for (bi = botlists[i]; bi; bi = bi->next)
    {
    	count++;
		mem += sizeof(*bi);
		mem += strlen(bi->nick)+1;
		mem += strlen(bi->user)+1;
		mem += strlen(bi->host)+1;
		mem += strlen(bi->real)+1;
    }
    }
	
    *nrec = count;
    *memuse = mem;
}

/*************************************************************************/
/*************************************************************************/

/* BotServ initialization. */

void bs_init(void)
{
}

/*************************************************************************/

/* Main BotServ routine. */

void botserv(const char *source, char *buf)
{
    char *cmd, *s;
    User *u = finduser(source);

    if (!u) {
	log("%s: user record for %s not found", s_BotServ, source);
	notice(s_BotServ, source,
		getstring((NickInfo *)NULL, USER_RECORD_NOT_FOUND));
	return;
    }

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;
    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_BotServ, source, "\1PING %s", s);
    } else if (skeleton) {
	notice_lang(s_BotServ, u, SERVICE_OFFLINE, s_BotServ);
    } else {
	run_cmd(s_BotServ, u, cmds, cmd);
    }

}

/*************************************************************************/

/* Handles all messages sent to bots. (Currently only answers to pings ;) */

void botmsgs(BotInfo *bi, const char *source, char *buf)
{
	char *cmd = strtok(buf, " ");
	char *s;
	
	if (!cmd) return;
	
	if (!stricmp(cmd, "\1PING")) {
		if (!(s = strtok(NULL, "")))
			s = "\1";
		notice(bi->nick, source, "\1PING %s", s);
	}
}

/*************************************************************************/

/* Handles all messages that are sent to registered channels where a 
 * bot is on.
 *
 */

void botchanmsgs(const char *source, ChannelInfo *ci, char *buf)
{
	int c;
	char *cmd;
	User *u = finduser(source);
	UserData *ud;
	
	if (!u) return;
	
	/* Now we can make kicker stuff. We try to order the checks
	 * from the fastest one to the slowest one, since there's
	 * no need to process other kickers if an user is kicked before
	 * the last kicker check.
	 *
	 * But FIRST we check whether the user is protected in any
	 * way.
	 */
	 
	if (buf && !check_access(u, ci, CA_NOKICK) && (!(ci->botflags & BS_DONTKICKOPS) || !is_chanop(u->nick, ci->c->name)) && (!(ci->botflags & BS_DONTKICKVOICES) || !is_voiced(u->nick, ci->c->name)))
	{
		/* Bolds kicker */
		if ((ci->botflags & BS_KICK_BOLDS) && strchr(buf, 2))
		{
			check_ban(ci, u, TTB_BOLDS);
			bot_kick(ci, u, BOT_REASON_BOLD);
			return;	
		}
		
		/* Color kicker */
		if ((ci->botflags & BS_KICK_COLORS) && strchr(buf, 3))
		{
			check_ban(ci, u, TTB_COLORS);
			bot_kick(ci, u, BOT_REASON_COLOR);
			return;
		}

		/* Reverses kicker */
		if ((ci->botflags & BS_KICK_REVERSES) && strchr(buf, 22))
		{
			check_ban(ci, u, TTB_REVERSES);
			bot_kick(ci, u, BOT_REASON_REVERSE);
			return;	
		}
		
		/* Underlines kicker */
		if ((ci->botflags & BS_KICK_UNDERLINES) && strchr(buf, 31))
		{
			check_ban(ci, u, TTB_UNDERLINES);
			bot_kick(ci, u, BOT_REASON_UNDERLINE);	
			return;
		}	
		
		/* Caps kicker */
		if ((ci->botflags & BS_KICK_CAPS) && ((c = strlen(buf)) >= ci->capsmin))
		{
			int i = 0;
			char *s = buf;
			
			do {
				if (isupper(*s)) i++;
			} while (*s++);
			  
			if (i >= ci->capsmin && i*100/c>=ci->capspercent)
			{
				check_ban(ci, u, TTB_CAPS);
				bot_kick(ci, u, BOT_REASON_CAPS);
				return;	
			}	
		}
		
		/* Bad words kicker */
		if (ci->botflags & BS_KICK_BADWORDS)
		{
			int i;
			int mustkick = 0;
			BadWord *bw;
			
			for (i = 0, bw = ci->badwords; i < ci->bwcount; i++, bw++)
			{
				if (!bw->in_use) continue;
				
				if (bw->type == BW_ANY && stristr(buf, bw->word)) mustkick = 1;
				else if (bw->type == BW_SINGLE) {
					int len = strlen(bw->word);
					
					if (!stricmp(buf, bw->word)) {
						mustkick = 1;
					/* two next if are quite odd isn't it? =) */
					} else if ((strchr(buf, ' ') == buf+len) && (stristr(buf, bw->word) == buf)) {
						mustkick = 1;
					} else if ((strrchr(buf, ' ') == buf+strlen(buf)-len-1) && (stristr(buf, bw->word) == buf+strlen(buf)-len)) {
						mustkick = 1;
					} else {
						char *wordbuf = scalloc(len+3, 1);
						
						wordbuf[0] = ' ';
						wordbuf[len+1] = ' ';
						wordbuf[len+2] = '\0';
						memcpy(wordbuf+1, bw->word, len);
						
						if (stristr(buf, wordbuf)) mustkick = 1;
					}
				} else if (bw->type == BW_START) {
					int len = strlen(bw->word);
					
					if (!strnicmp(buf, bw->word, len)) { 
						mustkick = 1;
					} else {
						char *wordbuf = scalloc(len+2, 1);
						
						memcpy(wordbuf+1, bw->word, len);
						wordbuf[0] = ' ';
						wordbuf[len+1] = '\0';
						
						if (stristr(buf, wordbuf)) mustkick = 1;
						
						free(wordbuf);
					}
				} else if (bw->type == BW_END) {
					int len = strlen(bw->word);
					
					if (!strnicmp(buf+strlen(buf)-len, bw->word, len)) {
						mustkick = 1; 
					} else {
						char *wordbuf = scalloc(len+2, 1);
						
						memcpy(wordbuf, bw->word, len);
						wordbuf[len] = ' ';
						wordbuf[len+1] = '\0';
						
						if (stristr(buf, wordbuf)) mustkick = 1;	
						
						free(wordbuf);
					}
				}
				
				if (mustkick)
				{
					check_ban(ci, u, TTB_BADWORDS);
					bot_kick(ci, u, BOT_REASON_BADWORD, bw->word);
					return;
				}
			} 
		}
		
		/* Flood kicker */
		if (ci->botflags & BS_KICK_FLOOD)
		{
			time_t now = time(NULL);
			
			ud = get_user_data(ci->c, u);
			if (!ud) return;
			
			if (now - ud->last_start > ci->floodsecs) 
			{
				ud->last_start = time(NULL);
				ud->lines = 0;
			}
			
			ud->lines++;
			if (ud->lines >= ci->floodlines)
			{
				check_ban(ci, u, TTB_FLOOD);
				bot_kick(ci, u, BOT_REASON_FLOOD);
				return;	
			}
		}
		
		/* Repeat kicker */
		if (ci->botflags & BS_KICK_REPEAT)
		{
			ud = get_user_data(ci->c, u);
			if (!ud) return;
			
			if (ud->lastline && stricmp(ud->lastline, buf)) {
				free(ud->lastline);
				ud->lastline = sstrdup(buf);
				ud->times = 0;	
			} else {
				if (!ud->lastline) ud->lastline = sstrdup(buf);
				ud->times++;
			}
			
			if (ud->times >= ci->repeattimes)
			{
				check_ban(ci, u, TTB_REPEAT);
				bot_kick(ci, u, BOT_REASON_REPEAT);
				return;	
			}
		}
	}
	
	/* Fantaisist commands */
	
	if (buf && (ci->botflags & BS_FANTASY) && check_access(u, ci, CA_FANTASIA)) {
		cmd = strtok(buf, " ");
		
		if (cmd) {
			if (!stricmp(cmd, "!deop")) {
				char *target = strtok(NULL, " ");
				
				if (!target && check_access(u, ci, CA_OPDEOPME))
					bot_raw_mode(ci, "-o", u->nick);
				else if (target && check_access(u, ci, CA_OPDEOP))
					bot_raw_mode(ci, "-o", target);	
#ifdef IRC_UNREAL
			} else if (!stricmp(cmd, "!dehalfop")) {
				char *target = strtok(NULL, " ");
				
				if (!target && check_access(u, ci, CA_HALFOPME))
					bot_raw_mode(ci, "-h", u->nick);
				else if (target && check_access(u, ci, CA_HALFOP))
					bot_raw_mode(ci, "-h", target);	
			} else if (!stricmp(cmd, "!deprotect")) {
				char *target = strtok(NULL, " ");
				
				if (!target && check_access(u, ci, CA_PROTECTME))
					bot_raw_mode(ci, "-a", u->nick);
				else if (target && check_access(u, ci, CA_PROTECT))
					bot_raw_mode(ci, "-a", target);	
#endif
			} else if (!stricmp(cmd, "!devoice")) {
				char *target = strtok(NULL, " ");
				
				if (!target && check_access(u, ci, CA_VOICEME))
					bot_raw_mode(ci, "-v", u->nick);
				else if (target && check_access(u, ci, CA_VOICE))
					bot_raw_mode(ci, "-v", target);
#ifdef IRC_UNREAL
			} else if (!stricmp(cmd, "!halfop")) {
				char *target = strtok(NULL, " ");
				
				if (!target && check_access(u, ci, CA_HALFOPME))
					bot_raw_mode(ci, "+h", u->nick);
				else if (target && check_access(u, ci, CA_HALFOP))
					bot_raw_mode(ci, "+h", target);	
#endif
			/* Looking for some secrets in the source code uh? :)
			   OK, here's one.. !kb has an alias named !kickban.
			   Wonderful isn't it? :) */
			} else if ((!stricmp(cmd, "!kb") || !stricmp(cmd, "!kickban")) && check_access(u, ci, CA_OPDEOP)) {
				char *target = strtok(NULL, " ");
				char *reason = strtok(NULL, "");
				
				if (!target) {
					bot_raw_ban(ci, u->nick);
					bot_raw_kick(ci, u->nick, "Requested");
				} else {
					if (!stricmp(target, ci->bi->nick))
					{
						bot_raw_ban(ci, u->nick);
						bot_raw_kick(ci, u->nick, "Oops!");
					} else {
						bot_raw_ban(ci, target);
						if (!reason)
							bot_raw_kick(ci, target, "Requested");
						else
							bot_raw_kick(ci, target, reason);	
					}
				}
			} else if (!stricmp(cmd, "!kick")) {
				char *target = strtok(NULL, " ");
				char *reason = strtok(NULL, "");
				
				if (!target && check_access(u, ci, CA_KICKME))
					bot_raw_kick(ci, u->nick, "Requested");
				else if (target && check_access(u, ci, CA_KICK)) {
					if (!stricmp(target, ci->bi->nick))
						bot_raw_kick(ci, u->nick, "Oops!");
					else if (!reason)
						bot_raw_kick(ci, target, "Requested");
					else
						bot_raw_kick(ci, target, reason);	
				}
			} else if (!stricmp(cmd, "!op")) {
				char *target = strtok(NULL, " ");
				
				if (!target && check_access(u, ci, CA_OPDEOPME))
					bot_raw_mode(ci, "+o", u->nick);
				else if (target && check_access(u, ci, CA_OPDEOP))
					bot_raw_mode(ci, "+o", target);	
#ifdef IRC_UNREAL
			} else if (!stricmp(cmd, "!protect")) {
				char *target = strtok(NULL, " ");
				
				if (!target && check_access(u, ci, CA_PROTECTME))
					bot_raw_mode(ci, "+a", u->nick);
				else if (target && check_access(u, ci, CA_PROTECT))
					bot_raw_mode(ci, "+a", target);	
#endif
			} else if (!stricmp(cmd, "!seen")) {
				char *target = strtok(NULL, " ");
				char buf[BUFSIZE];
				
				if (target) {
					User *u2;
					NickInfo *ni;
					ChanAccess *access;
					
					if (!stricmp(ci->bi->nick, target))
					{
						/* If we look for the bot */
						snprintf(buf, sizeof(buf), getstring(u->ni, BOT_SEEN_BOT), u->nick);
						send_cmd(ci->bi->nick, "PRIVMSG %s :%s", ci->name, buf);
					} else if (!(ni = findnick(target))) {
						/* If the nick is not registered */
						snprintf(buf, sizeof(buf), getstring(u->ni, BOT_SEEN_UNKNOWN), target);
						send_cmd(ci->bi->nick, "PRIVMSG %s :%s", ci->name, buf);
					} else if ((u2 = ni_on_chan(ci->c, getlink(ni)))) {
						/* If the nick we're looking for is on the channel,
						 * there are three possibilities: it's yourself,
						 * it's the nick we look for, it's a link of the
						 * nick we look for.
						 */
						if (!stricmp(u->nick, u2->nick)) 
						snprintf(buf, sizeof(buf), getstring(u->ni, BOT_SEEN_YOU), u->nick);
						else if (!stricmp(u2->nick, target))
						snprintf(buf, sizeof(buf), getstring(u->ni, BOT_SEEN_ON_CHANNEL), u2->nick);
						else
						snprintf(buf, sizeof(buf), getstring(u->ni, BOT_SEEN_ON_CHANNEL_AS), target, u2->nick);
						send_cmd(ci->bi->nick, "PRIVMSG %s :%s", ci->name, buf);
					} else if ((access = get_access_entry(getlink(ni), ci))) {
						/* User is on the access list but not present actually.
						   Special case: if access->last_seen is 0 it's that we
						   never seen the user.
						 */
						if (access->last_seen) {
							char durastr[192];
							duration(u->ni, durastr, sizeof(durastr), time(NULL) - access->last_seen);
							snprintf(buf, sizeof(buf), getstring(u->ni, BOT_SEEN_ON), target, durastr);
						} else {
							snprintf(buf, sizeof(buf), getstring(u->ni, BOT_SEEN_NEVER), target);
						}
						send_cmd(ci->bi->nick, "PRIVMSG %s :%s", ci->name, buf);
					} else {
						/* All other cases */
						snprintf(buf, sizeof(buf), getstring(u->ni, BOT_SEEN_UNKNOWN), target);
						send_cmd(ci->bi->nick, "PRIVMSG %s :%s", ci->name, buf);
					}
				}
				
			} else if (!stricmp(cmd, "!unban") && check_access(u, ci, CA_OPDEOP)) {
				char *target = strtok(NULL, " ");
				
				if (!target)
					bot_raw_unban(ci, u->nick);
				else
					bot_raw_unban(ci, target);
			} else if (!stricmp(cmd, "!voice")) {
				char *target = strtok(NULL, " ");
				
				if (!target && check_access(u, ci, CA_VOICEME))
					bot_raw_mode(ci, "+v", u->nick);
				else if (target && check_access(u, ci, CA_VOICE))
					bot_raw_mode(ci, "+v", target);	
			}
		}	
	}
	
	/* We try to answer to ping. */
	  
	cmd = strtok(buf, " ");
	if (cmd && !stricmp(cmd, "\1PING"))
	{
		char *s;
		if (!(s = strtok(NULL, "")))
			s = "\1";
		notice(ci->bi->nick, source, "\1PING %s", s);
	}
}

/*************************************************************************/

/* Load/save data files. */


#define SAFE(x) do {					\
    if ((x) < 0) {					\
	if (!forceload)					\
	    fatal("Read error on %s", BotDBName);	\
	failed = 1;					\
	break;						\
    }							\
} while (0)

void load_bs_dbase(void)
{
    dbFILE *f;
    int c, ver;
    int16 tmp16;
    int32 tmp32;
    BotInfo *bi;
    int failed = 0;

    if (!(f = open_db(s_BotServ, BotDBName, "r", BOT_VERSION)))
	return;

    switch (ver = get_file_version(f)) {

      case 9:
      case 8:

	while ((c = getc_db(f)) != 0) {
		char *s;
		
		if (c != 1)
		    fatal("Invalid format in %s %d", BotDBName, c);
			
		SAFE(read_string(&s, f));
		bi = makebot(s);
		free(s);	
		SAFE(read_string(&bi->user, f));
		SAFE(read_string(&bi->host, f));
		SAFE(read_string(&bi->real, f));
		SAFE(read_int32(&tmp32, f));
		bi->created = tmp32;
		SAFE(read_int16(&tmp16, f));
		bi->chancount = tmp16;
	} 	

	break;

      default:
	fatal("Unsupported version number (%d) on %s", ver, BotDBName);

    } /* switch (version) */

    close_db(f);
}

#undef SAFE

/*************************************************************************/

#define SAFE(x) do {						\
    if ((x) < 0) {						\
	restore_db(f);						\
	log_perror("Write error on %s", BotDBName);		\
	if (time(NULL) - lastwarn > WarningTimeout) {		\
	    wallops(NULL, "Write error on %s: %s", BotDBName,	\
			strerror(errno));			\
	    lastwarn = time(NULL);				\
	}							\
	return;							\
    }								\
} while (0)

void save_bs_dbase(void)
{
    dbFILE *f;
    BotInfo *bi;
    static time_t lastwarn = 0;
    int i;

    if (!(f = open_db(s_BotServ, BotDBName, "w", BOT_VERSION)))
	return;

	for (i = 0; i < 256; i++) {
    
    for (bi = botlists[i]; bi; bi = bi->next)
    {	
	SAFE(write_int8(1, f));
    	SAFE(write_string(bi->nick, f));	
	SAFE(write_string(bi->user, f));
	SAFE(write_string(bi->host, f));
	SAFE(write_string(bi->real, f));
	SAFE(write_int32(bi->created, f));
	SAFE(write_int16(bi->chancount, f));
	}
	
	}
    
    SAFE(write_int8(0, f));

    close_db(f);
}

#undef SAFE

/*************************************************************************/

/* Inserts a bot in the bot list. I can't be much explicit mh? */ 

static void insert_bot(BotInfo *bi)
{   
    /* bi->prev = NULL;
    bi->next = botlist;
    if (bi->next) bi->next->prev = bi;
    botlist = bi; */
    
    BotInfo *ptr, *prev;
    
    for (prev = NULL, ptr = botlists[tolower(*bi->nick)];
			ptr != NULL && stricmp(ptr->nick, bi->nick) < 0;
			prev = ptr, ptr = ptr->next)
	;
	bi->prev = prev;
    bi->next = ptr;
    if (!prev)
	botlists[tolower(*bi->nick)] = bi;
    else
	prev->next = bi;
    if (ptr)
	ptr->prev = bi;
}

/*************************************************************************/

static BotInfo *makebot(char *nick)
{
    BotInfo *bi;

    bi = scalloc(sizeof(BotInfo), 1);
    bi->nick = sstrdup(nick);
    bi->lastmsg = time(NULL);
    insert_bot(bi);
    nbots++;
    return bi;
}

/*************************************************************************/

static void change_bot_nick(BotInfo *bi, char *newnick) 
{
	if (bi->next)
	bi->next->prev = bi->prev;
    if (bi->prev)
	bi->prev->next = bi->next;
    else
    botlists[tolower(*bi->nick)] = bi->next;
	
	if (bi->nick) free(bi->nick);
	bi->nick = sstrdup(newnick);
	
	insert_bot(bi);
}

/*************************************************************************/

static int delbot(BotInfo *bi)
{
    cs_remove_bot(bi);
	
    if (bi->next)
	bi->next->prev = bi->prev;
    if (bi->prev)
	bi->prev->next = bi->next;
    else
    botlists[tolower(*bi->nick)] = bi->next;
    	
    nbots--;
    	
    free(bi->nick);
    free(bi->user);
    free(bi->host);
    free(bi->real);
    
    free(bi);
    
    return 1;
}

/*************************************************************************/

BotInfo *findbot(char *nick)
{
    BotInfo *bi;
    
    if (!nick || !*nick) return NULL;
    
    for (bi = botlists[tolower(*nick)]; bi; bi = bi->next)
    	if (!stricmp(nick, bi->nick)) return bi;
    
    return NULL;
}

/*************************************************************************/

/* Unassign a bot from a channel. Assumes u, ci and ci->bi are not NULL */

static void unassign(User *u, ChannelInfo *ci) 
{
	if (usercount(ci->c) >= BSMinUsers) 
		send_cmd(ci->bi->nick, "PART %s :UNASSIGN from %s", ci->name, u->nick);
	ci->bi->chancount--;
	ci->bi = NULL;
}

/*************************************************************************/

/* Returns ban data associated with an user if it exists, allocates it
   otherwise. */

static BanData *get_ban_data(Channel *c, User *u)
{
	char mask[BUFSIZE];
	BanData *bd, *next;
	time_t now = time(NULL);
	
	if (!c || !u) return NULL;
	
	snprintf(mask, sizeof(mask), "%s@%s", u->username, u->host);
	
	for (bd = c->bd; bd; bd = next) {
		if (now - bd->last_use > BSKeepData)
		{
			if (bd->next)
				bd->next->prev = bd->prev;
			if (bd->prev)
				bd->prev->next = bd->next;
			else
				c->bd = bd->next;
			if (bd->mask) free(bd->mask);
			next = bd->next;
			free(bd);
			continue;
		}
		if (!stricmp(bd->mask, mask)) 
		{
			bd->last_use = now;
			return bd;
		}
		next = bd->next;
	}
		
	/* If we fall here it is that we haven't found the record */
	bd = scalloc(sizeof(BanData), 1);
	bd->mask = sstrdup(mask);
	bd->last_use = now;
	
	bd->prev = NULL;
        bd->next = c->bd;
        if (bd->next) bd->next->prev = bd;
        c->bd = bd;
	
	return bd;
}

/*************************************************************************/

/* Returns BotServ data associated with an user on a given channel. 
 * Allocates it if necessary.
 */

static UserData *get_user_data(Channel *c, User *u)
{
	struct c_userlist *user;
	
	if (!c || !u) return NULL;
	
	for (user = c->users; user; user = user->next) {
		if (user->user == u) {
			if (user->ud) 
			{
				time_t now = time(NULL);
				
				/* Checks whether data is obsolete */
				if (now - user->ud->last_use > BSKeepData)
				{
					if (user->ud->lastline) free(user->ud->lastline);
					/* We should not free and realloc, but reset to 0
					   instead. */
					memset(user->ud, 0, sizeof(UserData));
					user->ud->last_use = now;
				}
				
				return user->ud;	
			} else {
				user->ud = scalloc(sizeof(UserData), 1);
				user->ud->last_use = time(NULL);
				return user->ud;	
			}	
		}
	}
	
	return NULL;
}

/*************************************************************************/

/* Makes the bot join a channel and op himself. */

void bot_join(ChannelInfo *ci)
{	
	int i;
	
	if (!ci || !ci->c || !ci->bi) return;	
	
	if (BSSmartJoin)
	{	
		/* We check for bans */
		int count = ci->c->bancount;
		if (count) {
		char botmask[BUFSIZE];
		char **bans = scalloc(sizeof(char *) * count, 1);
		char *av[3];
		
		memcpy(bans, ci->c->bans, sizeof(char *) * count);
		snprintf(botmask, sizeof(botmask), "%s!%s@%s", ci->bi->nick, ci->bi->user, ci->bi->host);

		av[0] = ci->c->name;
		av[1] = sstrdup("-b");
		for (i = 0; i < count; i++) {
	    	if (match_wild_nocase(ci->c->bans[i], botmask)) {
				send_cmd(ci->bi->nick, "MODE %s -b %s", ci->name, bans[i]);
				av[2] = sstrdup(bans[i]);
				do_cmode(ci->bi->nick, 3, av);
				free(av[2]);
	    	}
		}
		free(av[1]);
		free(bans);
		}
		
		/* Should we be invited? */
		if ((ci->c->mode & CMODE_I) || (ci->c->limit && usercount(ci->c) >= ci->c->limit))
			send_cmd(NULL, "NOTICE @%s :%s invited %s into the channel.", ci->c->name, ci->bi->nick, ci->bi->nick);
	}
	
#ifdef IRC_BAHAMUT
	send_cmd(ci->bi->nick, "SJOIN %ld %s", ci->c->creation_time, ci->name);
#else
	send_cmd(ci->bi->nick, "JOIN %s", ci->name);
#endif

#ifdef IRC_UNREAL
	send_cmd(ci->bi->nick, "MODE %s +oa %s %s", ci->name, ci->bi->nick, ci->bi->nick);
#else
	send_cmd(ci->bi->nick, "MODE %s +o %s", ci->name, ci->bi->nick);
#endif
}

/*************************************************************************/

/* This makes the bot rejoin all channel he is on when he gets killed
 * or changed.
 */
 
void bot_rejoin_all(BotInfo *bi)
{
	int i;
	ChannelInfo *ci;
	
	for (i = 0; i < 256; i++)
	{
		for (ci = chanlists[i]; ci; ci = ci->next) {
	 		if (ci->bi == bi && (usercount(ci->c) >= BSMinUsers))
	    			bot_join(ci);
	    	}
	}
}

/*************************************************************************/

/* This makes a ban if the user has to have one. In every cases it increments
   the kick count for the user. */

static void check_ban(ChannelInfo *ci, User *u, int ttbtype)
{
	BanData *bd = get_ban_data(ci->c, u);
	
	if (!bd) return;
	
	bd->ttb[ttbtype]++;
	if (bd->ttb[ttbtype] == ci->ttb[ttbtype])
	{
		char *av[3];
		char mask[BUFSIZE];
		
		bd->ttb[ttbtype] = 0;
		
		av[0] = ci->name;
		av[1] = sstrdup("+b");
		get_idealban(ci, u, mask, sizeof(mask));
		av[2] = mask;
		send_cmd(ci->bi->nick, "MODE %s +b %s", av[0], av[2]);
		do_cmode(ci->bi->nick, 3, av);
		free(av[1]);	
	} 
}

/*************************************************************************/

/* This makes a bot kick an user. Works somewhat like notice_lang in fact ;) */

static void bot_kick(ChannelInfo *ci, User *u, int message, ...)
{
    va_list args;
    char buf[1024];
    const char *fmt;
    char *av[3];

    if (!ci || !ci->bi || !ci->c || !u) return;
    
    va_start(args, message);
    fmt = getstring(u->ni, message);
    if (!fmt) return;
    vsnprintf(buf, sizeof(buf), fmt, args);
    
    av[0] = ci->name;
    av[1] = u->nick;
    av[2] = buf;
    send_cmd(ci->bi->nick, "KICK %s %s :%s", av[0], av[1], av[2]);
    do_kick(ci->bi->nick, 3, av);
}

/*************************************************************************/

/* Makes a simple ban. */

static void bot_raw_ban(ChannelInfo *ci, char *nick)
{
	User *u = finduser(nick);
	char *av[3];
	char mask[BUFSIZE];
		
	if (!u) return;
	
#ifdef IRC_UNREAL
	if (is_founder(u, ci) || is_protected(nick, ci->name)) return;
#endif
		
	av[0] = ci->name;
	av[1] = sstrdup("+b");
	get_idealban(ci, u, mask, sizeof(mask));
	av[2] = mask;
	send_cmd(ci->bi->nick, "MODE %s +b %s", av[0], av[2]);
	do_cmode(ci->bi->nick, 3, av);
	free(av[1]);	 		
}

/*************************************************************************/

/* Makes a kick with a "dynamic" reason ;) */

static void bot_raw_kick(ChannelInfo *ci, char *nick, char *reason)
{
    char *av[3];
#ifdef IRC_UNREAL
	User *u = finduser(nick);
#endif
    
    if (!ci || !ci->bi || !ci->c || !nick || !is_on_chan(nick, ci->name)) return;
    
#ifdef IRC_UNREAL
	if (is_founder(u, ci) || is_protected(nick, ci->name)) return;
#endif
    
    av[0] = ci->name;
    av[1] = nick;
    av[2] = (reason ? reason : ci->bi->nick);
    send_cmd(ci->bi->nick, "KICK %s %s :%s", av[0], av[1], av[2]);
    do_kick(ci->bi->nick, 3, av);
}

/*************************************************************************/

/* Makes a mode operation on a channel for a nick */

static void bot_raw_mode(ChannelInfo *ci, char *mode, char *nick)
{
	char *av[3];
#ifdef IRC_UNREAL
	User *u = finduser(nick);
#endif
	
	if (!ci || !ci->c || !ci->bi || !mode || !nick) return;
	if (!is_on_chan(nick, ci->name)) return;
	
#ifdef IRC_UNREAL
	if (!stricmp(mode, "-o") && (is_founder(u, ci) || is_protected(nick, ci->name))) return;
#endif
	
	av[0] = ci->name;
	av[1] = mode;
	av[2] = nick;
	
	send_cmd(ci->bi->nick, "MODE %s %s %s", av[0], av[1], av[2]);
	do_cmode(ci->bi->nick, 3, av); 
}

/*************************************************************************/

/* Removes all bans for a nick on a channel */

static void bot_raw_unban(ChannelInfo *ci, char *nick)
{
	int count, i;
	char mask[BUFSIZE];
	char *av[3], **bans;
	User *u;
	
	if (!ci || !ci->c || !ci->bi || !nick) return;
	if (!(u = finduser(nick))) return;
	
	snprintf(mask, sizeof(mask), "%s!%s@%s", u->nick, u->username, u->host);
	
	av[0] = ci->name;
	av[1] = sstrdup("-b");
	
	count = ci->c->bancount;
	bans =scalloc(sizeof(char *) * count, 1);
	memcpy(bans, ci->c->bans, sizeof(char *) * count);
	
	for (i = 0; i < count; i++) {
	    if (match_wild_nocase(bans[i], mask)) {
			send_cmd(ci->bi->nick, "MODE %s -b %s", ci->name, bans[i]);
			av[2] = bans[i];
			do_cmode(ci->bi->nick, 3, av);
		}
	}
	free(bans);
	free(av[1]);
}

/*************************************************************************/
/*************************************************************************/

static void do_help(User *u)
{
    char *cmd = strtok(NULL, "");

    if (!cmd) {
	notice_help(s_BotServ, u, BOT_HELP, BSMinUsers);
	if (is_services_oper(u))
	    notice_help(s_BotServ, u, BOT_SERVADMIN_HELP);
    } else {
	help_cmd(s_BotServ, u, cmds, cmd);
    }
}

/*************************************************************************/

static void do_bot(User *u)
{
    BotInfo *bi;
    char *cmd = strtok(NULL, " ");
    
    if (!cmd)
    	syntax_error(s_BotServ, u, "BOT", BOT_BOT_SYNTAX);
    else if (!stricmp(cmd, "ADD"))
    {
    	char *nick = strtok(NULL, " ");	
    	char *user = strtok(NULL, " ");
    	char *host = strtok(NULL, " ");
    	char *real = strtok(NULL, "");
    	
    	if (!nick || !user || !host || !real)
    		syntax_error(s_BotServ, u, "BOT", BOT_BOT_SYNTAX);
    	else if (readonly)
    		notice_lang(s_BotServ, u, BOT_BOT_READONLY);
    	else if (findbot(nick))
    		notice_lang(s_BotServ, u, BOT_BOT_ALREADY_EXISTS, nick);
    	else
    	{
    		User *u2;
    		NickInfo *ni;
    		
    		bi = makebot(nick);
    		if (!bi)
    		{
    			notice_lang(s_BotServ, u, BOT_BOT_CREATION_FAILED);
    			return;	
    		}
    		
    		bi->user = sstrdup(user);
    		bi->host = sstrdup(host);
    		bi->real = sstrdup(real);
    		bi->created = time(NULL);
    		bi->chancount = 0;
    		
    		/* We check whether the nick is registered, and drop it if so */
    		if ((ni = findnick(nick))) 
    			delnick(ni);
    		
    		/* We check whether user with this nick is online, and kill it if so */
    		if ((u2 = finduser(nick)))
    			kill_user(s_BotServ, u2->nick, "This nick is now used by Services");
    			
    		/* We make the bot online, ready to serve */
    		NEWNICK(bi->nick, bi->user, bi->host, bi->real, "+r", 1);
    		
    		notice_lang(s_BotServ, u, BOT_BOT_ADDED, bi->nick, bi->user, bi->host, bi->real);
    	}
    }
    else if (!stricmp(cmd, "CHANGE"))
    {
    	char *oldnick = strtok(NULL, " ");
    	char *nick = strtok(NULL, " ");	
    	char *user = strtok(NULL, " ");
    	char *host = strtok(NULL, " ");
    	char *real = strtok(NULL, "");
    	
    	if (!oldnick || !nick)
    		syntax_error(s_BotServ, u, "BOT", BOT_BOT_SYNTAX);
    	else if (readonly)
    		notice_lang(s_BotServ, u, BOT_BOT_READONLY);
    	else if (!(bi = findbot(oldnick)))
    		notice_lang(s_BotServ, u, BOT_DOES_NOT_EXIST, oldnick);
    	else
    	{
    		User *u2;
    		NickInfo *ni;
    		
    		/* Checks whether there *are* changes.
    		 * Case sensitive because we may want to change just the case.
    		 * And we must finally check that the nick is not already
    		 * taken by another bot.
    		 */
    		if (!strcmp(bi->nick, nick) 
    				&& ((user) ? !strcmp(bi->user, user) : 1)
    				&& ((host) ? !strcmp(bi->host, host) : 1)
    				&& ((real) ? !strcmp(bi->real, real) : 1))
    		{
    			notice_lang(s_BotServ, u, BOT_BOT_ANY_CHANGES);
    			return;
    		}
    		
    		if (stricmp(bi->nick, nick) && findbot(nick))
    		{
    			notice_lang(s_BotServ, u, BOT_BOT_ALREADY_EXISTS, nick);
    			return;
    		}
    		
    		if (stricmp(bi->nick, nick))
    		{
    			/* The new nick is really different, so we remove the Q line for 
    		   	   the old nick. */
    		   	send_cmd(NULL, "UNSQLINE %s", bi->nick);
    			
    			/* We check whether the nick is registered, and drop it if so */
    			if ((ni = findnick(nick))) 
    				delnick(ni);
    		
    			/* We check whether user with this nick is online, and kill it if so */
    			if ((u2 = finduser(nick)))
    				kill_user(s_BotServ, u2->nick, "This nick is now used by Services");
    		}
    		
    		if (strcmp(nick, bi->nick)) 
    			change_bot_nick(bi, nick);
    			
    		if (user && strcmp(user, bi->user))
    		{
    			free(bi->user);
    			bi->user = sstrdup(user);
    		}
    		if (host && strcmp(host, bi->host))
    		{
    			free(bi->host);
    			bi->host = sstrdup(host);
    		}
    		if (real && strcmp(real, bi->real))
    		{
    			free(bi->real);
    			bi->real = sstrdup(real);
    		}
    		
    		/* If only the nick changes, we just make the bot change his nick,
    		   else we must make it quit and rejoin. */
    		if (!user) 
    			send_cmd(oldnick, "NICK %s", bi->nick);
    		else
    		{
    			send_cmd(oldnick, "QUIT :Quit: Be right back");
    			NEWNICK(bi->nick, bi->user, bi->host, bi->real, "+r", 1);
    			bot_rejoin_all(bi);
    		}
    		
    		notice_lang(s_BotServ, u, BOT_BOT_CHANGED, oldnick, bi->nick, bi->user, bi->host, bi->real);
    	}
    }
    else if (!stricmp(cmd, "DEL"))
    {
    	char *nick = strtok(NULL, " ");
    	
    	if (!nick)
    		syntax_error(s_BotServ, u, "BOT", BOT_BOT_SYNTAX);
    	else if (readonly)
    		notice_lang(s_BotServ, u, BOT_BOT_READONLY);
    	else if (!(bi = findbot(nick)))
    		notice_lang(s_BotServ, u, BOT_DOES_NOT_EXIST, nick);
    	else
    	{
    		send_cmd(bi->nick, "QUIT :Quit: Help! I'm being deleted by %s!", u->nick);
    		send_cmd(NULL, "UNSQLINE %s", bi->nick);
    		delbot(bi);
    		
    		notice_lang(s_BotServ, u, BOT_BOT_DELETED, nick);
    	}
    }
    else if (!stricmp(cmd, "LIST")) do_botlist(u);
    else syntax_error(s_BotServ, u, "BOT", BOT_BOT_SYNTAX);
}

/*************************************************************************/

static void do_botlist(User *u)
{
	int i;
	BotInfo *bi;
	
	if (!nbots)
	{
		notice_lang(s_BotServ, u, BOT_BOTLIST_EMPTY);
		return;
	}
	
	notice_lang(s_BotServ, u, BOT_BOTLIST_HEADER);
		
	for (i = 0; i < 256; i++)
		for (bi = botlists[i]; bi; bi = bi->next)
			notice(s_BotServ, u->nick, "   %-15s  (%s@%s)", bi->nick, bi->user, bi->host);	
}

/*************************************************************************/

static void do_assign(User *u)
{
	char *chan = strtok(NULL, " ");
	char *nick = strtok(NULL, " ");
	BotInfo *bi;
	ChannelInfo *ci;
	
	if (readonly)
		notice_lang(s_BotServ, u, BOT_ASSIGN_READONLY); 
	else if (!chan || !nick)
		syntax_error(s_BotServ, u, "ASSIGN", BOT_ASSIGN_SYNTAX);
	else if (!(ci = cs_findchan(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    	else if (ci->flags & CI_VERBOTEN)
		notice_lang(s_BotServ, u, CHAN_X_FORBIDDEN, chan);
	else if (ci->bi)
		notice_lang(s_BotServ, u, BOT_ASSIGN_ALREADY, chan);
	else if (!(bi = findbot(nick)))
		notice_lang(s_BotServ, u, BOT_DOES_NOT_EXIST, nick);
	else if ((ci->botflags & BS_NOBOT) || !check_access(u, ci, CA_ASSIGN))
		notice_lang(s_BotServ, u, PERMISSION_DENIED);
	else
	{
		ci->bi = bi;
		bi->chancount++;
		if (ci->c && usercount(ci->c) >= BSMinUsers) 
			bot_join(ci);	
		notice_lang(s_BotServ, u, BOT_ASSIGN_ASSIGNED, bi->nick, ci->name);
	}
}

/*************************************************************************/

static void do_unassign(User *u)
{
	char *chan = strtok(NULL, " ");
	ChannelInfo *ci;
	
	if (readonly)
		notice_lang(s_BotServ, u, BOT_ASSIGN_READONLY); 
	else if (!chan)
		syntax_error(s_BotServ, u, "UNASSIGN", BOT_UNASSIGN_SYNTAX);
	else if (!(ci = cs_findchan(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    else if (ci->flags & CI_VERBOTEN)
		notice_lang(s_BotServ, u, CHAN_X_FORBIDDEN, chan);
	else if (!check_access(u, ci, CA_ASSIGN))
		notice_lang(s_BotServ, u, ACCESS_DENIED);
	else
	{
		if (ci->bi) unassign(u, ci);
		notice_lang(s_BotServ, u, BOT_UNASSIGN_UNASSIGNED, ci->name);
	}
}

/*************************************************************************/

static void send_bot_channels(User *u, BotInfo *bi)
{
	int i;
	ChannelInfo *ci;
	char buf[307], *end;
	
	*buf = 0;
	end = buf;
	
	for (i = 0; i < 256; i++) {
		for (ci = chanlists[i]; ci; ci = ci->next) {
	    	if (ci->bi == bi) {
	    		if (strlen(buf) + strlen(ci->name) > 300) {
	    			notice(s_BotServ, u->nick, buf);
	    			*buf = 0;
					end = buf;
				}
				end += snprintf(end, sizeof(buf)-(end-buf), " %s ", ci->name);
	    	} 
		}
    }	
    
    if (*buf) notice(s_BotServ, u->nick, buf);
}

static void do_info(User *u)
{
	BotInfo *bi;
	ChannelInfo *ci;
	char *query = strtok(NULL, " ");
	
	int need_comma = 0, is_servadmin = is_services_admin(u);
	char buf[BUFSIZE], *end;
	const char *commastr = getstring(u->ni, COMMA_SPACE);
		
	if (!query)
		syntax_error(s_BotServ, u, "INFO", BOT_INFO_SYNTAX);
	else if ((bi = findbot(query)))
	{
		char buf[BUFSIZE];
		struct tm *tm;
		
		notice_lang(s_BotServ, u, BOT_INFO_BOT_HEADER, bi->nick);
		notice_lang(s_BotServ, u, BOT_INFO_BOT_MASK, bi->user, bi->host);
		notice_lang(s_BotServ, u, BOT_INFO_BOT_REALNAME, bi->real);
		tm = localtime(&bi->created);
		strftime_lang(buf, sizeof(buf), u, STRFTIME_DATE_TIME_FORMAT, tm);
		notice_lang(s_BotServ, u, BOT_INFO_BOT_CREATED, buf);
		notice_lang(s_BotServ, u, BOT_INFO_BOT_USAGE, bi->chancount);
		
		if (is_services_admin(u))
			send_bot_channels(u, bi);
		
	} else if ((ci = cs_findchan(query))) {
		if (!is_servadmin && !is_founder(u, ci)) {
			notice_lang(s_BotServ, u, PERMISSION_DENIED);
			return;	
		}
		
		notice_lang(s_BotServ, u, BOT_INFO_CHAN_HEADER, ci->name);
		if (ci->bi) notice_lang(s_BotServ, u, BOT_INFO_CHAN_BOT, ci->bi->nick);
		else notice_lang(s_BotServ, u, BOT_INFO_CHAN_BOT_NONE);
		
		if (ci->botflags & BS_KICK_BADWORDS)
		{
			if (ci->ttb[TTB_BADWORDS]) notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BADWORDS_BAN, getstring(u->ni, BOT_INFO_ACTIVE), ci->ttb[TTB_BADWORDS]);
			else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BADWORDS, getstring(u->ni, BOT_INFO_ACTIVE));
		} else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BADWORDS, getstring(u->ni, BOT_INFO_INACTIVE));
		if (ci->botflags & BS_KICK_BOLDS)
		{
			if (ci->ttb[TTB_BOLDS]) notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BOLDS_BAN, getstring(u->ni, BOT_INFO_ACTIVE), ci->ttb[TTB_BOLDS]);
			else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BOLDS, getstring(u->ni, BOT_INFO_ACTIVE));
		} else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_BOLDS, getstring(u->ni, BOT_INFO_INACTIVE));
		if (ci->botflags & BS_KICK_CAPS)
		{
			if (ci->ttb[TTB_CAPS]) notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_CAPS_BAN, getstring(u->ni, BOT_INFO_ACTIVE), ci->ttb[TTB_CAPS], ci->capsmin, ci->capspercent);
			else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_CAPS_ON, getstring(u->ni, BOT_INFO_ACTIVE), ci->capsmin, ci->capspercent);
		} else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_CAPS_OFF, getstring(u->ni, BOT_INFO_INACTIVE));
		if (ci->botflags & BS_KICK_COLORS)
		{
			if (ci->ttb[TTB_COLORS]) notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_COLORS_BAN, getstring(u->ni, BOT_INFO_ACTIVE), ci->ttb[TTB_COLORS]);
			else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_COLORS, getstring(u->ni, BOT_INFO_ACTIVE));
		} else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_COLORS, getstring(u->ni, BOT_INFO_INACTIVE));
		if (ci->botflags & BS_KICK_FLOOD)
		{
			if (ci->ttb[TTB_FLOOD]) notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_FLOOD_BAN, getstring(u->ni, BOT_INFO_ACTIVE), ci->ttb[TTB_FLOOD], ci->floodlines, ci->floodsecs);
			else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_FLOOD_ON, getstring(u->ni, BOT_INFO_ACTIVE), ci->floodlines, ci->floodsecs);
		} else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_FLOOD_OFF, getstring(u->ni, BOT_INFO_INACTIVE));
		if (ci->botflags & BS_KICK_REPEAT)
		{
			if (ci->ttb[TTB_REPEAT]) notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REPEAT_BAN, getstring(u->ni, BOT_INFO_ACTIVE), ci->ttb[TTB_REPEAT], ci->repeattimes);
			else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REPEAT_ON, getstring(u->ni, BOT_INFO_ACTIVE), ci->repeattimes);
		} else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REPEAT_OFF, getstring(u->ni, BOT_INFO_INACTIVE));
		if (ci->botflags & BS_KICK_REVERSES)
		{
			if (ci->ttb[TTB_REVERSES]) notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REVERSES_BAN, getstring(u->ni, BOT_INFO_ACTIVE), ci->ttb[TTB_REVERSES]);
			else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REVERSES, getstring(u->ni, BOT_INFO_ACTIVE));
		} else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_REVERSES, getstring(u->ni, BOT_INFO_INACTIVE));
		if (ci->botflags & BS_KICK_UNDERLINES)
		{
			if (ci->ttb[TTB_UNDERLINES]) notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_UNDERLINES_BAN, getstring(u->ni, BOT_INFO_ACTIVE), ci->ttb[TTB_UNDERLINES]);
			else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_UNDERLINES, getstring(u->ni, BOT_INFO_ACTIVE));
		} else notice_lang(s_BotServ, u, BOT_INFO_CHAN_KICK_UNDERLINES, getstring(u->ni, BOT_INFO_INACTIVE));
		
		end = buf;
		*end = 0;
		if (ci->botflags & BS_DONTKICKOPS) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s",
				getstring(u->ni, BOT_INFO_OPT_DONTKICKOPS));
	    	need_comma = 1;
		}
		if (ci->botflags & BS_DONTKICKVOICES) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
				need_comma ? commastr : "",
				getstring(u->ni, BOT_INFO_OPT_DONTKICKVOICES));
	    	need_comma = 1;
		}
		if (ci->botflags & BS_FANTASY) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
				need_comma ? commastr : "",
				getstring(u->ni, BOT_INFO_OPT_FANTASY));
	    	need_comma = 1;
		}
		if (ci->botflags & BS_GREET) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
				need_comma ? commastr : "",
				getstring(u->ni, BOT_INFO_OPT_GREET));
	    	need_comma = 1;
		}
		if (ci->botflags & BS_NOBOT) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
				need_comma ? commastr : "",
				getstring(u->ni, BOT_INFO_OPT_NOBOT));
	    	need_comma = 1;
		}
		if (ci->botflags & BS_SYMBIOSIS) {
	    	end += snprintf(end, sizeof(buf)-(end-buf), "%s%s",
				need_comma ? commastr : "",
				getstring(u->ni, BOT_INFO_OPT_SYMBIOSIS));
	    	need_comma = 1;
		}
		notice_lang(s_BotServ, u, BOT_INFO_CHAN_OPTIONS,
			*buf ? buf : getstring(u->ni, BOT_INFO_OPT_NONE));
			
	} else notice_lang(s_BotServ, u, BOT_INFO_NOT_FOUND, query);
}

/*************************************************************************/

static void do_set(User *u)
{
	char *chan = strtok(NULL, " ");
	char *option = strtok(NULL, " ");
	char *value = strtok(NULL, " ");
	int is_servadmin = is_services_admin(u);
	
	ChannelInfo *ci;
	
	if (readonly)
		notice_lang(s_BotServ, u, BOT_SET_DISABLED);
	else if (!chan || !option || !value)
		syntax_error(s_BotServ, u, "SET", BOT_SET_SYNTAX);
	else if (!(ci = cs_findchan(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    else if (ci->flags & CI_VERBOTEN)
		notice_lang(s_BotServ, u, CHAN_X_FORBIDDEN, chan);
    else if (!is_servadmin && !check_access(u, ci, CA_SET))
		notice_lang(s_BotServ, u, ACCESS_DENIED);
	else
	{
		if (!stricmp(option, "DONTKICKOPS")) {
			if (!stricmp(value, "ON")) {
				ci->botflags |= BS_DONTKICKOPS;
				notice_lang(s_BotServ, u, BOT_SET_DONTKICKOPS_ON, ci->name);
			} else if (!stricmp(value, "OFF")) {
				ci->botflags &= ~BS_DONTKICKOPS;
				notice_lang(s_BotServ, u, BOT_SET_DONTKICKOPS_OFF, ci->name);	
			} else {
				syntax_error(s_BotServ, u, "SET DONTKICKOPS", BOT_SET_DONTKICKOPS_SYNTAX);	
			}
		} else if (!stricmp(option, "DONTKICKVOICES")) {
			if (!stricmp(value, "ON")) {
				ci->botflags |= BS_DONTKICKVOICES;
				notice_lang(s_BotServ, u, BOT_SET_DONTKICKVOICES_ON, ci->name);
			} else if (!stricmp(value, "OFF")) {
				ci->botflags &= ~BS_DONTKICKVOICES;
				notice_lang(s_BotServ, u, BOT_SET_DONTKICKVOICES_OFF, ci->name);	
			} else {
				syntax_error(s_BotServ, u, "SET DONTKICKVOICES", BOT_SET_DONTKICKVOICES_SYNTAX);	
			}
		} else if (!stricmp(option, "FANTASY")) {
			if (!stricmp(value, "ON")) {
				ci->botflags |= BS_FANTASY;
				notice_lang(s_BotServ, u, BOT_SET_FANTASY_ON, ci->name);
			} else if (!stricmp(value, "OFF")) {
				ci->botflags &= ~BS_FANTASY;
				notice_lang(s_BotServ, u, BOT_SET_FANTASY_OFF, ci->name);	
			} else {
				syntax_error(s_BotServ, u, "SET FANTASY", BOT_SET_FANTASY_SYNTAX);	
			}
		} else if (!stricmp(option, "GREET")) {
			if (!stricmp(value, "ON")) {
				ci->botflags |= BS_GREET;
				notice_lang(s_BotServ, u, BOT_SET_GREET_ON, ci->name);
			} else if (!stricmp(value, "OFF")) {
				ci->botflags &= ~BS_GREET;
				notice_lang(s_BotServ, u, BOT_SET_GREET_OFF, ci->name);	
			} else {
				syntax_error(s_BotServ, u, "SET GREET", BOT_SET_GREET_SYNTAX);	
			}
		} else if (is_servadmin && !stricmp(option, "NOBOT")) {
			if (!stricmp(value, "ON")) {
				ci->botflags |= BS_NOBOT;
				if (ci->bi) unassign(u, ci);
				notice_lang(s_BotServ, u, BOT_SET_NOBOT_ON, ci->name);
			} else if (!stricmp(value, "OFF")) {
				ci->botflags &= ~BS_NOBOT;
				notice_lang(s_BotServ, u, BOT_SET_NOBOT_OFF, ci->name);	
			} else {
				syntax_error(s_BotServ, u, "SET NOBOT", BOT_SET_NOBOT_SYNTAX);	
			}
		} else if (!stricmp(option, "SYMBIOSIS")) {
			if (!stricmp(value, "ON")) {
				ci->botflags |= BS_SYMBIOSIS;
				notice_lang(s_BotServ, u, BOT_SET_SYMBIOSIS_ON, ci->name);
			} else if (!stricmp(value, "OFF")) {
				ci->botflags &= ~BS_SYMBIOSIS;
				notice_lang(s_BotServ, u, BOT_SET_SYMBIOSIS_OFF, ci->name);
			} else {
				syntax_error(s_BotServ, u, "SET SYMBIOSIS", BOT_SET_SYMBIOSIS_SYNTAX);	
			}
		} else {
			notice_help(s_BotServ, u, BOT_SET_UNKNOWN, option);
		}
	}
}

/*************************************************************************/

static void do_kickcmd(User *u)
{
	char *chan = strtok(NULL, " ");
	char *option = strtok(NULL, " ");
	char *value = strtok(NULL, " ");
	char *ttb = strtok(NULL, " ");
	
	ChannelInfo *ci;
	
	if (readonly)
		notice_lang(s_BotServ, u, BOT_KICK_DISABLED);
	else if (!chan || !option || !value)
		syntax_error(s_BotServ, u, "KICK", BOT_KICK_SYNTAX);
	else if (stricmp(value, "ON") && stricmp(value, "OFF"))
		syntax_error(s_BotServ, u, "KICK", BOT_KICK_SYNTAX);
	else if (!(ci = cs_findchan(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    	else if (ci->flags & CI_VERBOTEN)
		notice_lang(s_BotServ, u, CHAN_X_FORBIDDEN, chan);
    	else if (!is_services_admin(u) && !check_access(u, ci, CA_SET))
		notice_lang(s_BotServ, u, ACCESS_DENIED);
	else
	{
		if (!stricmp(option, "BADWORDS")) {
			if (!stricmp(value, "ON")) {
				if (ttb) {
					ci->ttb[TTB_BADWORDS] = atol(ttb);
					if (ci->ttb[TTB_BADWORDS] < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, ttb);
						return;
					}
				} else ci->ttb[TTB_BADWORDS] = 0;
				ci->botflags |= BS_KICK_BADWORDS;
				if (ci->ttb[TTB_BADWORDS]) notice_lang(s_BotServ, u, BOT_KICK_BADWORDS_ON_BAN, ci->ttb[TTB_BADWORDS]);
				else notice_lang(s_BotServ, u, BOT_KICK_BADWORDS_ON);
			} else {
				ci->botflags &= ~BS_KICK_BADWORDS;
				notice_lang(s_BotServ, u, BOT_KICK_BADWORDS_OFF);	
			}	
		} else if (!stricmp(option, "BOLDS")) {
			if (!stricmp(value, "ON")) {
				if (ttb) {
					ci->ttb[TTB_BOLDS] = atol(ttb);
					if (ci->ttb[TTB_BOLDS] < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, ttb);
						return;
					}
				} else ci->ttb[TTB_BOLDS] = 0;
				ci->botflags |= BS_KICK_BOLDS;
				if (ci->ttb[TTB_BOLDS]) notice_lang(s_BotServ, u, BOT_KICK_BOLDS_ON_BAN, ci->ttb[TTB_BOLDS]);
				else notice_lang(s_BotServ, u, BOT_KICK_BOLDS_ON);
			} else {
				ci->botflags &= ~BS_KICK_BOLDS;
				notice_lang(s_BotServ, u, BOT_KICK_BOLDS_OFF);	
			}	
		} else if (!stricmp(option, "CAPS")) {
			if (!stricmp(value, "ON")) {
				char *min = strtok(NULL, " ");
				char *percent = strtok(NULL, " ");
				
				if (ttb) {
					ci->ttb[TTB_CAPS] = atol(ttb);
					if (ci->ttb[TTB_CAPS] < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, ttb);
						return;
					}
				} else ci->ttb[TTB_CAPS] = 0;
				
				if (!min) ci->capsmin = 10;
				else ci->capsmin = atol(min);
				if (ci->capsmin < 1) ci->capsmin = 10; 
				
				if (!percent) ci->capspercent = 25;
				else ci->capspercent = atol(percent);
				if (ci->capspercent < 1 || ci->capspercent > 100) ci->capspercent = 25; 
				
				ci->botflags |= BS_KICK_CAPS;
				if (ci->ttb[TTB_CAPS]) notice_lang(s_BotServ, u, BOT_KICK_CAPS_ON_BAN, ci->capsmin, ci->capspercent, ci->ttb[TTB_CAPS]);
				else notice_lang(s_BotServ, u, BOT_KICK_CAPS_ON, ci->capsmin, ci->capspercent);
			} else {
				ci->botflags &= ~BS_KICK_CAPS;
				notice_lang(s_BotServ, u, BOT_KICK_CAPS_OFF);	
			}	
		} else if (!stricmp(option, "COLORS")) {
			if (!stricmp(value, "ON")) {
				if (ttb) {
					ci->ttb[TTB_COLORS] = atol(ttb);
					if (ci->ttb[TTB_COLORS] < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, ttb);
						return;
					}
				} else ci->ttb[TTB_COLORS] = 0;
				ci->botflags |= BS_KICK_COLORS;
				if (ci->ttb[TTB_COLORS]) notice_lang(s_BotServ, u, BOT_KICK_COLORS_ON_BAN, ci->ttb[TTB_COLORS]);
				else notice_lang(s_BotServ, u, BOT_KICK_COLORS_ON);
			} else {
				ci->botflags &= ~BS_KICK_COLORS;
				notice_lang(s_BotServ, u, BOT_KICK_COLORS_OFF);	
			}
		} else if (!stricmp(option, "FLOOD")) {
			if (!stricmp(value, "ON")) {
				char *lines = strtok(NULL, " ");
				char *secs = strtok(NULL, " ");
				
				if (ttb) {
					ci->ttb[TTB_FLOOD] = atol(ttb);
					if (ci->ttb[TTB_FLOOD] < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, ttb);
						return;
					}
				} else ci->ttb[TTB_FLOOD] = 0;
				
				if (!lines) ci->floodlines = 6;
				else ci->floodlines = atol(lines);
				if (ci->floodlines < 2) ci->floodlines = 6; 
				
				if (!secs) ci->floodsecs = 10;
				else ci->floodsecs = atol(secs);
				if (ci->floodsecs < 1 || ci->floodsecs > BSKeepData) ci->floodsecs = 10; 
				
				ci->botflags |= BS_KICK_FLOOD;
				if (ci->ttb[TTB_FLOOD]) notice_lang(s_BotServ, u, BOT_KICK_FLOOD_ON_BAN, ci->floodlines, ci->floodsecs, ci->ttb[TTB_FLOOD]);
				else notice_lang(s_BotServ, u, BOT_KICK_FLOOD_ON, ci->floodlines, ci->floodsecs);
			} else {
				ci->botflags &= ~BS_KICK_FLOOD;
				notice_lang(s_BotServ, u, BOT_KICK_FLOOD_OFF);	
			}
		} else if (!stricmp(option, "REPEAT")) {
			if (!stricmp(value, "ON")) {
				char *times = strtok(NULL, " ");
				
				if (ttb) {
					ci->ttb[TTB_REPEAT] = atol(ttb);
					if (ci->ttb[TTB_REPEAT] < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, ttb);
						return;
					}
				} else ci->ttb[TTB_REPEAT] = 0;
				
				if (!times) ci->repeattimes = 3;
				else ci->repeattimes = atol(times);
				if (ci->repeattimes < 2) ci->repeattimes = 3; 
				
				ci->botflags |= BS_KICK_REPEAT;
				if (ci->ttb[TTB_REPEAT]) notice_lang(s_BotServ, u, BOT_KICK_REPEAT_ON_BAN, ci->repeattimes, ci->ttb[TTB_REPEAT]);
				else notice_lang(s_BotServ, u, BOT_KICK_REPEAT_ON, ci->repeattimes);
			} else {
				ci->botflags &= ~BS_KICK_REPEAT;
				notice_lang(s_BotServ, u, BOT_KICK_REPEAT_OFF);	
			}
		} else if (!stricmp(option, "REVERSES")) {
			if (!stricmp(value, "ON")) {
				if (ttb) {
					ci->ttb[TTB_REVERSES] = atol(ttb);
					if (ci->ttb[TTB_REVERSES] < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, ttb);
						return;
					}
				} else ci->ttb[TTB_REVERSES] = 0;
				ci->botflags |= BS_KICK_REVERSES;
				if (ci->ttb[TTB_REVERSES]) notice_lang(s_BotServ, u, BOT_KICK_REVERSES_ON_BAN, ci->ttb[TTB_REVERSES]);
				else notice_lang(s_BotServ, u, BOT_KICK_REVERSES_ON);
			} else {
				ci->botflags &= ~BS_KICK_REVERSES;
				notice_lang(s_BotServ, u, BOT_KICK_REVERSES_OFF);	
			}
		} else if (!stricmp(option, "UNDERLINES")) {
			if (!stricmp(value, "ON")) {
				if (ttb) {
					ci->ttb[TTB_UNDERLINES] = atol(ttb);
					if (ci->ttb[TTB_UNDERLINES] < 0) {
						notice_lang(s_BotServ, u, BOT_KICK_BAD_TTB, ttb);
						return;
					}
				}  else ci->ttb[TTB_UNDERLINES] = 0;
				ci->botflags |= BS_KICK_UNDERLINES;
				if (ci->ttb[TTB_UNDERLINES]) notice_lang(s_BotServ, u, BOT_KICK_UNDERLINES_ON_BAN, ci->ttb[TTB_UNDERLINES]);
				else notice_lang(s_BotServ, u, BOT_KICK_UNDERLINES_ON);
			} else {
				ci->botflags &= ~BS_KICK_UNDERLINES;
				notice_lang(s_BotServ, u, BOT_KICK_UNDERLINES_OFF);	
			}
		} else notice_help(s_BotServ, u, BOT_KICK_UNKNOWN, option);
	}
}

/*************************************************************************/

static int badwords_del_callback(User *u, int num, va_list args)
{
    BadWord *bw;
    ChannelInfo *ci = va_arg(args, ChannelInfo *);
    int *last = va_arg(args, int *);
    if (num < 1 || num > ci->bwcount)
	return 0;
    *last = num;
    
    bw = &ci->badwords[num-1];
    if (bw->word) 
	free(bw->word);
    bw->word = NULL;
    bw->in_use = 0;
    
    return 1;
}

static int badwords_list(User *u, int index, ChannelInfo *ci, int *sent_header)
{
    BadWord *bw = &ci->badwords[index];
    
    if (!bw->in_use)
	return 0;
    if (!*sent_header) {
	notice_lang(s_BotServ, u, BOT_BADWORDS_LIST_HEADER, ci->name);
	*sent_header = 1;
    }
  
    notice_lang(s_BotServ, u, BOT_BADWORDS_LIST_FORMAT, index+1, bw->word,
    			((bw->type == BW_SINGLE) ? "(SINGLE)" : 
    			((bw->type == BW_START) ? "(START)" : 
    			((bw->type == BW_END) ? "(END)" : "")))
    			);
    return 1;
}

static int badwords_list_callback(User *u, int num, va_list args)
{
    ChannelInfo *ci = va_arg(args, ChannelInfo *);
    int *sent_header = va_arg(args, int *);
    if (num < 1 || num > ci->bwcount)
	return 0;
    return badwords_list(u, num-1, ci, sent_header);
}

static void do_badwords(User *u)
{
    char *chan = strtok(NULL, " ");
    char *cmd  = strtok(NULL, " ");
    char *word = strtok(NULL, "");
    ChannelInfo *ci;
    BadWord *bw;
    
    int i;
    int need_args = (cmd && (!stricmp(cmd, "LIST") || !stricmp(cmd, "CLEAR")));

    if (!cmd || (need_args ? 0 : !word)) {
	syntax_error(s_BotServ, u, "BADWORDS", BOT_BADWORDS_SYNTAX);
    } else if (!(ci = cs_findchan(chan))) {
	notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    } else if (ci->flags & CI_VERBOTEN) {
	notice_lang(s_BotServ, u, CHAN_X_FORBIDDEN, chan);
    } else if (!check_access(u, ci, CA_BADWORDS) && (!need_args || !is_services_admin(u))) {
	notice_lang(s_BotServ, u, ACCESS_DENIED);
    } else if (stricmp(cmd, "ADD") == 0) {
    	
    char *opt, *pos;
    int type = BW_ANY;
    	
	if (readonly) {
	    notice_lang(s_BotServ, u, BOT_BADWORDS_DISABLED);
	    return;
	}
	
	pos = strrchr(word, ' ');
	if (pos) {
		opt = pos+1;
		if (*opt) {
			if (!stricmp(opt, "SINGLE")) type = BW_SINGLE;
			else if (!stricmp(opt, "START")) type = BW_START;
			else if (!stricmp(opt, "END")) type = BW_END;
			if (type != BW_ANY) *pos = 0;
		}
	}

	for (bw = ci->badwords, i = 0; i < ci->bwcount; bw++, i++) {
	    if (bw->word && !stricmp(bw->word, word)) {
		notice_lang(s_BotServ, u, BOT_BADWORDS_ALREADY_EXISTS, bw->word, ci->name);
		return;
	    }
	}
	
	for (i = 0; i < ci->bwcount; i++) {
	    if (!ci->badwords[i].in_use)
		break;
	}
	if (i == ci->bwcount) {
	    if (i < BSBadWordsMax) {
		ci->bwcount++;
		ci->badwords = srealloc(ci->badwords, sizeof(BadWord) * ci->bwcount);
	    } else {
		notice_lang(s_BotServ, u, BOT_BADWORDS_REACHED_LIMIT, BSBadWordsMax);
		return;
	    }
	}
	bw = &ci->badwords[i];
	bw->in_use = 1;
	bw->word = sstrdup(word);
	bw->type = type;
	
	notice_lang(s_BotServ, u, BOT_BADWORDS_ADDED, bw->word, ci->name);

    } else if (stricmp(cmd, "DEL") == 0) {

	if (readonly) {
	    notice_lang(s_BotServ, u, BOT_BADWORDS_DISABLED);
	    return;
	}

	/* Special case: is it a number/list?  Only do search if it isn't. */
	if (isdigit(*word) && strspn(word, "1234567890,-") == strlen(word)) {
	    int count, deleted, last = -1;
	    deleted = process_numlist(word, &count, badwords_del_callback, u,
					ci, &last);
	    if (!deleted) {
		if (count == 1) {
		    notice_lang(s_BotServ, u, BOT_BADWORDS_NO_SUCH_ENTRY,
				last, ci->name);
		} else {
		    notice_lang(s_BotServ, u, BOT_BADWORDS_NO_MATCH, ci->name);
		}
	    } else if (deleted == 1) {
		notice_lang(s_BotServ, u, BOT_BADWORDS_DELETED_ONE, ci->name);
	    } else {
		notice_lang(s_BotServ, u, BOT_BADWORDS_DELETED_SEVERAL,
				deleted, ci->name);
	    }
	} else {
	    for (i = 0; i < ci->bwcount; i++) {
		if (ci->badwords[i].in_use && !stricmp(ci->badwords[i].word, word))
		    break;
	    }
	    if (i == ci->bwcount) {
		notice_lang(s_BotServ, u, BOT_BADWORDS_NOT_FOUND, word, chan);
		return;
	    }
	    bw = &ci->badwords[i];
	    notice_lang(s_BotServ, u, BOT_BADWORDS_DELETED, bw->word, ci->name);
	    if (bw->word) 
	        free(bw->word);
	    bw->word = NULL;
	    bw->in_use = 0;
	}

    } else if (stricmp(cmd, "LIST") == 0) {
	int sent_header = 0;

	if (ci->bwcount == 0) {
	    notice_lang(s_BotServ, u, BOT_BADWORDS_LIST_EMPTY, chan);
	    return;
	}
	if (word && strspn(word, "1234567890,-") == strlen(word)) {
	    process_numlist(word, NULL, badwords_list_callback, u, ci,
								&sent_header);
	} else {
	    for (i = 0; i < ci->bwcount; i++) {
	    if (!(ci->badwords[i].in_use)) continue; 
		if (word && ci->badwords[i].word
			 && !match_wild_nocase(word, ci->badwords[i].word))
		    continue;
		badwords_list(u, i, ci, &sent_header);
	    }
	}
	if (!sent_header)
	    notice_lang(s_BotServ, u, BOT_BADWORDS_NO_MATCH, chan);
	    
	} else if (stricmp(cmd, "CLEAR") == 0) {

		if (readonly) {
	    	notice_lang(s_BotServ, u, BOT_BADWORDS_DISABLED);
	    	return;
		}

	    for (i = 0; i < ci->bwcount; i++)
			if (ci->badwords[i].word) free(ci->badwords[i].word);
	    
	    free(ci->badwords);
	    ci->badwords = NULL;
	    ci->bwcount = 0;

		notice_lang(s_BotServ, u, BOT_BADWORDS_CLEAR);

    } else {
	syntax_error(s_BotServ, u, "BADWORDS", BOT_BADWORDS_SYNTAX);
    }
}

/*************************************************************************/

static void do_say(User *u)
{
	ChannelInfo *ci;
	
	char *chan = strtok(NULL, " ");
	char *text = strtok(NULL, "");
	
	if (!chan || !text)
		syntax_error(s_BotServ, u, "SAY", BOT_SAY_SYNTAX);
	else if (!(ci = cs_findchan(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    else if (ci->flags & CI_VERBOTEN)
		notice_lang(s_BotServ, u, CHAN_X_FORBIDDEN, chan);
	else if (!ci->bi)
		notice_help(s_BotServ, u, BOT_NOT_ASSIGNED);
	else if (!ci->c || usercount(ci->c) < BSMinUsers)
		notice_lang(s_BotServ, u, BOT_NOT_ON_CHANNEL, ci->name);
    else if (!check_access(u, ci, CA_SAY))
		notice_lang(s_BotServ, u, ACCESS_DENIED);
	else {
		send_cmd(ci->bi->nick, "PRIVMSG %s :%s", ci->name, text);	
		ci->bi->lastmsg = time(NULL);
	}
}

/*************************************************************************/

static void do_act(User *u)
{
	ChannelInfo *ci;
	
	char *chan = strtok(NULL, " ");
	char *text = strtok(NULL, "");
	
	if (!chan || !text)
		syntax_error(s_BotServ, u, "ACT", BOT_ACT_SYNTAX);
	else if (!(ci = cs_findchan(chan)))
		notice_lang(s_BotServ, u, CHAN_X_NOT_REGISTERED, chan);
    else if (ci->flags & CI_VERBOTEN)
		notice_lang(s_BotServ, u, CHAN_X_FORBIDDEN, chan);
	else if (!ci->bi)
		notice_help(s_BotServ, u, BOT_NOT_ASSIGNED);
	else if (!ci->c || usercount(ci->c) < BSMinUsers)
		notice_lang(s_BotServ, u, BOT_NOT_ON_CHANNEL, ci->name);
    else if (!check_access(u, ci, CA_SAY))
		notice_lang(s_BotServ, u, ACCESS_DENIED);
	else {
		send_cmd(ci->bi->nick, "PRIVMSG %s :%cACTION %s%c", ci->name, 1, text, 1);
		ci->bi->lastmsg = time(NULL);	
	}
}

/*************************************************************************/
